"""Concept factory."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Union, cast, get_args

from . import Language
from .concept import Concept, ConceptId, ConceptIds, ConceptRelation, RelatedConceptIds, RootConceptIds
from .grammar import GrammaticalCategory
from .iana_language_subtag_registry import ALL_LANGUAGES
from .label import Labels, label_factory, meaning_factory

ConceptIdListOrString = ConceptId | list[ConceptId]
ConceptIdDictOrListOrString = dict[Language, ConceptIdListOrString] | ConceptIdListOrString
ConceptAttribute = Literal["antonym", "answer", "answer-only", "example", "holonym", "hypernym", "involves", "roots"]
LeafConceptDict = dict[Language | ConceptAttribute, ConceptId | list[ConceptId] | ConceptIdDictOrListOrString | bool]
CompositeConceptDict = dict[
    GrammaticalCategory | ConceptAttribute, Union["CompositeConceptDict", LeafConceptDict, bool]
]
ConceptDict = LeafConceptDict | CompositeConceptDict


@dataclass(frozen=True)
class ConceptFactory:
    """Create concepts from the concept dict."""

    concept_id: ConceptId
    concept_dict: ConceptDict

    def create_concept(self, parent: ConceptId | None = None) -> Concept:
        """Create a concept from the concept_dict."""
        return Concept(
            self.concept_id,
            parent,
            self._constituent_concepts(),
            self._labels(),
            self._meanings(),
            self._related_concepts(),
            self._root_concepts(),
            self._answer_only(),
        )

    def _labels(self) -> dict[Language, Labels]:
        """Return the concept labels."""
        return {
            cast(Language, key): label_factory(cast(Language, key), cast(str | list[str], value))
            for key, value in self.concept_dict.items()
            if key in ALL_LANGUAGES
        }

    def _meanings(self) -> dict[Language, Labels]:
        """Return the concept meanings."""
        return {
            cast(Language, key): meaning_factory(cast(Language, key), cast(str | list[str], value))
            for key, value in self.concept_dict.items()
            if key in ALL_LANGUAGES
        }

    def _related_concepts(self) -> RelatedConceptIds:
        """Create the related concepts."""
        return {relation: self._related_concept_ids(relation) for relation in get_args(ConceptRelation)}

    def _answer_only(self) -> bool:
        """Return whether the concept is answer-only."""
        return bool(self.concept_dict.get("answer-only"))

    def _constituent_concepts(self) -> ConceptIds:
        """Create a constituent concept for each grammatical category."""
        return tuple(self._constituent_concept(category) for category in self._grammatical_categories())

    def _constituent_concept(self, category: GrammaticalCategory) -> ConceptId:
        """Create a constituent concept for the specified grammatical category."""
        constituent_concept_id = ConceptId(f"{self.concept_id}/{category}")
        concept_factory = self.__class__(constituent_concept_id, self._constituent_concept_dict(category))
        concept_factory.create_concept(self.concept_id)
        return constituent_concept_id

    def _constituent_concept_dict(self, category: GrammaticalCategory) -> ConceptDict:
        """Create a constituent concept dict that can be used to create a constituent concept."""
        antonym_concept_ids = self._related_concept_ids("antonym")
        antonyms_dict = dict(antonym=[ConceptId(f"{antonym}/{category}") for antonym in antonym_concept_ids])
        answer_concept_ids = self._related_concept_ids("answer")
        answers_dict = dict(answer=[ConceptId(f"{answer}/{category}") for answer in answer_concept_ids])
        roots_dict = cast(CompositeConceptDict, dict(roots=self._get_roots()))
        constituent_concept_dict = (
            cast(ConceptDict, cast(CompositeConceptDict, self.concept_dict)[category])
            | roots_dict
            | antonyms_dict
            | answers_dict
        )
        return cast(ConceptDict, constituent_concept_dict)

    def _grammatical_categories(self) -> tuple[GrammaticalCategory, ...]:
        """Retrieve the grammatical categories from the concept dict."""
        keys = self.concept_dict.keys()
        return tuple(cast(GrammaticalCategory, key) for key in keys if key in get_args(GrammaticalCategory))

    def _root_concepts(self) -> RootConceptIds:
        """Retrieve the roots from the concept dict."""
        roots = self._get_roots()
        if isinstance(roots, dict):  # Roots are not the same for all languages
            return {
                language: tuple(concept_ids) if isinstance(concept_ids, list) else (concept_ids,)
                for language, concept_ids in roots.items()
            }
        return tuple(roots) if isinstance(roots, list) else (roots,)

    def _get_roots(self) -> ConceptIdDictOrListOrString:
        """Get the roots from the concept dict."""
        return cast(ConceptIdDictOrListOrString, self.concept_dict.get("roots", {}))

    def _related_concept_ids(self, relation: ConceptAttribute) -> ConceptIds:
        """Return the ids of the related concept(s)."""
        related = cast(ConceptIdListOrString, self.concept_dict.get(relation, []))
        return tuple(related) if isinstance(related, list) else (related,)


def create_concept(concept_id: ConceptId, concept_dict: ConceptDict) -> Concept:
    """Create a concept from the concept dict."""
    return ConceptFactory(concept_id, concept_dict or cast(ConceptDict, {})).create_concept()



"""Filter concepts."""

from argparse import ArgumentParser

from .language import Language
from .language.concept import Concept, ConceptId


def filter_concepts(
    concepts: set[Concept],
    selected_concepts: list[ConceptId],
    language: Language,
    argument_parser: ArgumentParser,
) -> set[Concept]:
    """Filter the concepts by selected concepts."""
    if not selected_concepts:
        return concepts
    all_selected_concepts: set[Concept] = set()
    concepts_by_concept_id = {concept.concept_id: concept for concept in concepts}
    for concept_id in selected_concepts:
        if concept_id not in concepts_by_concept_id:
            argument_parser.error(f"Concept '{concept_id}' not found\n")
        all_selected_concepts.add(concepts_by_concept_id[concept_id])
    for concept in all_selected_concepts.copy():
        all_selected_concepts |= set(concept.get_related_concepts("hyponym") + concept.get_related_concepts("meronym"))
    for concept in all_selected_concepts.copy():
        all_selected_concepts |= set(
            concept.get_related_concepts("antonym")
            + concept.get_related_concepts("hyponym")
            + concept.get_related_concepts("meronym")
            + concept.get_related_concepts("involves")
            + concept.get_related_concepts("involved_by")
            + concept.compounds(language)
        )
    return all_selected_concepts



"""Labels."""

from __future__ import annotations

import re
from collections.abc import Sequence
from functools import cached_property
from typing import ClassVar, Final

from . import Language

SpellingAlternatives = dict[Language, dict[re.Pattern[str], str]]

END_OF_SENTENCE_PUNCTUATION = """?!'"."""


class Label(str):
    """Class representing labels for concepts."""

    __slots__ = ("__dict__", "language")  # Without adding __dict__ to slots @cached_property does not work

    # Labels can have one question note and multiple answer notes. The question note is shown before a quiz is
    # presented to the user. The answer notes are shown afterwards. The format is:
    # 'label;question note;answer note 1;answer note 2;...'
    # If the label itself ends with an asterisk it's a colloquial label, i.e. spoken language only.
    NOTE_SEP: Final = ";"
    COLLOQUIAL_POSTFIX: Final = "*"
    QUESTION_NOTE_INDEX: Final = 1
    ANSWER_NOTE_INDEX: Final = 2
    SPELLING_ALTERNATIVES_SEP: Final = "|"
    ALTERNATIVES_TO_GENERATE: ClassVar[SpellingAlternatives] = {}  # These are loaded upon start of the application

    def __new__(cls, _language: Language, value: str) -> Label:  # noqa: PYI034
        """Create an instance of the label."""
        return super().__new__(cls, value)  # This is needed because we overwrite an immutable type (str)

    def __init__(self, language: Language, _value: str) -> None:
        """Initialize the label."""
        self.language = language

    def __eq__(self, other: object) -> bool:
        """Return whether the labels are equal."""
        if isinstance(other, Label):
            return self.language == other.language and self.without_notes == other.without_notes
        return False

    def __ne__(self, other: object) -> bool:
        """Return whether the labels are not equal."""
        if isinstance(other, Label):
            return self.language != other.language or self.without_notes != other.without_notes
        return True

    @property
    def non_generated_spelling_alternatives(self) -> Labels:
        """Extract the spelling alternatives from the label."""
        return label_factory(self.language, self.without_notes.split(self.SPELLING_ALTERNATIVES_SEP))

    @cached_property
    def spelling_alternatives(self) -> Labels:
        """Extract the spelling alternatives from the label and generate additional spelling alternatives."""
        alternatives = self.non_generated_spelling_alternatives
        generated_alternatives = set()
        for alternative in alternatives:
            for pattern, replacement in self.ALTERNATIVES_TO_GENERATE.get(self.language, {}).items():
                if re.search(pattern, alternative):
                    generated_alternative = re.sub(pattern, replacement, alternative)
                    if alternative[0].isupper():
                        generated_alternative = generated_alternative[0].capitalize() + generated_alternative[1:]
                    generated_alternatives.add(generated_alternative)
        return alternatives + tuple(Label(self.language, alternative) for alternative in generated_alternatives)

    @property
    def question_note(self) -> str:
        """Return the label question note."""
        has_question_note = self.count(self.NOTE_SEP) >= self.QUESTION_NOTE_INDEX
        return self.split(self.NOTE_SEP)[self.QUESTION_NOTE_INDEX] if has_question_note else ""

    @property
    def answer_notes(self) -> Sequence[str]:
        """Return the label answer notes."""
        has_answer_notes = self.count(self.NOTE_SEP) >= self.ANSWER_NOTE_INDEX
        return self.split(self.NOTE_SEP)[self.ANSWER_NOTE_INDEX :] if has_answer_notes else ()

    @property
    def without_notes(self) -> str:
        """Return the label without the notes."""
        return self.split(self.NOTE_SEP)[0]

    @property
    def is_colloquial(self) -> bool:
        """Return whether this is a colloquial label."""
        return self.without_notes.endswith(self.COLLOQUIAL_POSTFIX)

    @property
    def is_complete_sentence(self) -> bool:
        """Return whether this is a complete sentence (starts with a capital and ends with punctuation)."""
        return self.without_notes[0].isupper() and (self.without_notes[-1] in END_OF_SENTENCE_PUNCTUATION)

    @property
    def pronounceable(self) -> str:
        """Return the label as text that can be sent to a speech synthesizer."""
        return self.without_notes.rstrip(self.COLLOQUIAL_POSTFIX).replace("'", "").replace("-", " ")


Labels = tuple[Label, ...]


def label_factory(language: Language, string: str | list[str]) -> Labels:
    """Instantiate the labels from a string or list of strings."""
    labels = string if isinstance(string, list) else [string]
    return tuple(Label(language, label) for label in labels if not label.startswith("("))


def meaning_factory(language: Language, string: str | list[str]) -> Labels:
    """Instantiate the meanings from a string or list of strings."""
    meanings = string if isinstance(string, list) else [string]
    return tuple(
        Label(language, meaning.removeprefix("(").removesuffix(")"))
        for meaning in meanings
        if not meaning.endswith(Label.COLLOQUIAL_POSTFIX)
    )



"""Loader."""

from argparse import ArgumentParser
from pathlib import Path

from ..metadata import NAME
from ..model.language.concept import Concept, ConceptId
from ..model.language.concept_factory import ConceptDict, create_concept
from .identifier_registry import IdentifierRegistry
from .json_file import load_json


class Loader:
    """Domain object loader."""

    def __init__(self, argument_parser: ArgumentParser) -> None:
        self.argument_parser = argument_parser
        self.concept_id_registry = IdentifierRegistry[ConceptId]("concept", argument_parser)

    def load(self, *files: Path) -> set[Concept]:
        """Load the domain objects from the files."""
        all_concepts = set()
        try:
            for file_path in files:
                concepts = self._parse_json(load_json(file_path))
                self.concept_id_registry.check_and_register_identifiers(self._concept_identifiers(concepts), file_path)
                all_concepts |= concepts
        except Exception as reason:  # noqa: BLE001
            self.argument_parser.error(f"{NAME} cannot read file {file_path}: {reason}.\n")
        return all_concepts

    def _parse_json(self, json: dict[ConceptId, ConceptDict]) -> set[Concept]:
        """Parse the domain objects from the JSON loaded from the domain object file."""
        return self._create_concepts(json)

    def _create_concepts(self, concept_dict: dict[ConceptId, ConceptDict]) -> set[Concept]:
        """Parse the concepts from the JSON."""
        return {create_concept(concept_key, concept_value) for concept_key, concept_value in concept_dict.items()}

    def _concept_identifiers(self, concepts: set[Concept]) -> tuple[ConceptId, ...]:
        """Return the identifiers of the concepts."""
        return tuple(concept.concept_id for concept in concepts)



"""Create a colored diff."""

from difflib import SequenceMatcher

from .dictionary import linkify
from .style import DELETED, INSERTED


def show_whitespace(text: str) -> str:
    """Make whitespace visible so it can be colored."""
    return "_" * len(text) if all(char.isspace() for char in text) else text


def inserted(new_text: str) -> str:
    """Return the annotated text."""
    return f"[{INSERTED}]{new_text}[/{INSERTED}]"


def deleted(old_text: str) -> str:
    """Return the annotated text."""
    return f"[{DELETED}]{show_whitespace(old_text)}[/{DELETED}]"


def colored_diff(old_text: str, new_text: str, min_ratio_for_diff: float = 0.6) -> str:
    """Return a colored string showing the diffs between old and new text."""
    matcher = SequenceMatcher(a=old_text.lower(), b=new_text.lower())
    if matcher.ratio() < min_ratio_for_diff:
        return inserted(linkify(new_text))
    result = ""
    for operator, old_start, old_end, new_start, new_end in matcher.get_opcodes():
        old_fragment, new_fragment = (old_text[old_start:old_end], new_text[new_start:new_end])
        if operator == "delete":
            result += deleted(old_fragment)
        elif operator == "insert":
            result += inserted(new_fragment)
        elif operator == "replace":
            if len(old_fragment) == len(new_fragment) == 1:
                result += deleted(old_fragment) + inserted(new_fragment)
            else:
                result += inserted(new_fragment)
        else:
            result += new_fragment
    return result



"""Meta data about this application."""

from importlib.metadata import metadata, version
from pathlib import Path
from typing import Final

import requests

from toisto.model.language import Language
from toisto.tools import first

_metadata = metadata("Toisto")
NAME: Final = _metadata["name"]
SUMMARY: Final = _metadata["summary"]
_homepage_url = _metadata["Project-URL"].split(", ")[1]
README_URL: Final = f"{_homepage_url}?tab=readme-ov-file#toisto"
CHANGELOG_URL: Final = first(_metadata.get_all("Project-URL", []), lambda url: "Changelog" in url).split(", ")[1]
TAGS_API_URL: Final = "https://api.github.com/repos/fniessink/toisto/tags"

VERSION: Final = version(NAME)
BUILT_IN_LANGUAGES: Final = [Language("en"), Language("fi"), Language("nl")]

# File locations
_data_folder = Path(__file__).parent.parent
CONCEPT_JSON_FILES: Final = sorted((_data_folder / "concepts").glob("**/*.json"))
_languages_folder = _data_folder / "languages"
LANGUAGES_FILE: Final = _languages_folder / "iana-language-subtag-registry.txt"
SPELLING_ALTERNATIVES_FILE: Final = _languages_folder / "spelling_alternatives.json"


def get_progress_filepath(target_language: Language) -> Path:
    """Return the filename of the progress file for the specified target language."""
    return Path.home() / f".{NAME.lower()}-progress-{target_language}.json"


def latest_version() -> str | None:
    """Return the latest version."""
    timeout: Final = 2
    try:
        response = requests.get(TAGS_API_URL, timeout=timeout)
        response.raise_for_status()  # We get a 403 if rate limited
        return str(response.json()[0]["name"])
    except requests.RequestException:
        return None



"""Grammar types."""

from typing import Literal

GrammaticalGender = Literal["female", "male", "neuter"]
GrammaticalNumber = Literal["infinitive", "singular", "plural"]
GrammaticalPerson = Literal["first person", "second person", "third person"]
DegreeOfComparison = Literal["positive degree", "comparative degree", "superlative degree"]
Diminutive = Literal["root", "diminutive"]
Tense = Literal["present tense", "past tense"]
GrammaticalMood = Literal["declarative", "interrogative", "imperative"]
GrammaticalPolarity = Literal["affirmative", "negative"]
Number = Literal["cardinal", "ordinal"]
Abbreviation = Literal["abbreviation", "full form"]
GrammaticalCategory = Literal[
    GrammaticalGender,
    GrammaticalNumber,
    GrammaticalPerson,
    DegreeOfComparison,
    Diminutive,
    Tense,
    GrammaticalMood,
    GrammaticalPolarity,
    Number,
    Abbreviation,
]



"""All languages as defined in the IANA language subtag registry."""

from typing import Final

from toisto.persistence.iana_language_subtag_registry import load_languages

IANA_LANGUAGE_SUBTAG_REGISTRY_URL: Final = "https://www.iana.org/assignments/language-subtag-registry"
ALL_LANGUAGES: Final = load_languages()



"""Concept classes."""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import ClassVar, Literal, NewType, cast, get_args

from ...tools import Registry
from . import Language
from .grammar import GrammaticalCategory
from .label import Labels

ConceptId = NewType("ConceptId", str)
ConceptIds = tuple[ConceptId, ...]
InvertedConceptRelation = Literal["hyponym", "involved_by", "meronym"]
RecursiveConceptRelation = Literal["holonym", "hypernym"]
NonInvertedConceptRelation = Literal[RecursiveConceptRelation, "antonym", "answer", "example", "involves"]
ConceptRelation = Literal[InvertedConceptRelation, NonInvertedConceptRelation]
RelatedConceptIds = dict[ConceptRelation, ConceptIds]
RootConceptIds = dict[Language, ConceptIds] | ConceptIds  # Tuple if all languages have the same roots


def inverted(relation: InvertedConceptRelation) -> ConceptRelation:
    """Return the inverted relation."""
    return cast(ConceptRelation, {"hyponym": "hypernym", "involved_by": "involves", "meronym": "holonym"}[relation])


@dataclass(frozen=True)
class Concept:
    """Class representing language concepts.

    A concept is either a composite or a leaf concept. Composite concepts have two or more constituent concepts,
    representing different grammatical categories, for example singular and plural forms. Leaf concepts have labels
    in different languages.

    Concepts can have following types of relations to other concepts:

    - The antonym relation is used to capture opposites. For example 'bad' has 'good' as antonym and 'good' has 'bad'
      as antonym.

    - The hypernym relation is used to capture "is a (kind of)" relations. Y is a hypernym of X if every X is a
      (kind of) Y. For example, canine is a hypernym of dog. The hypernym relation is transitive.

    - The holonym relation is used to capture "part of" relations. Y is a holonym of X if X is a part of Y.
      For example, sentences are holonym of words. The holonym relation is transitive.

    - The involves relation is used to link verbs with nouns.

    - The answer relation is used to specify possible answers to questions. Using the answer relation it is possible to
      specify that, for example, 'Do you like ice cream?' has the yes and no concepts as possible answers.

    - The examples relation is used to specify other concepts that exemplify the concept.

    - The roots relation is used to capture the relation between compound concepts and their roots. For example,
      the word 'blackboard' contains two roots, 'black' and 'board'. The concept for the compound concept refers to its
      roots with the roots attribute. The roots relation can also be used for sentences, in which case the individual
      words of a sentence are the roots.

      Note that this relationship is different from the other types of concept relations because the roots relationship
      is based on the label of the concept. It also means that the relationships can be different for different
      languages. For example, the Dutch label 'schoolbord' has different roots than the English equivalent 'blackboard'.

    NOTE: This class keeps track of the related concepts using their concept identifier (ConceptId) and only when
    the client asks for a concept is the concept instance looked up in the concept registry (Concept.instances). This
    prevents the need for a second pass after instantiating concepts from the concept files to create the relations.
    """

    concept_id: ConceptId
    _parent: ConceptId | None
    _constituents: ConceptIds
    _labels: dict[Language, Labels]
    _meanings: dict[Language, Labels]
    _related_concepts: RelatedConceptIds
    _roots: RootConceptIds
    answer_only: bool

    instances: ClassVar[Registry[ConceptId, Concept]] = Registry[ConceptId, "Concept"]()

    def __post_init__(self) -> None:
        """Add the concept to the concept registry."""
        self.instances.add_item(self.concept_id, self)

    def __hash__(self) -> int:
        """Return the concept hash."""
        return hash(self.concept_id)

    def get_related_concepts(self, relation: ConceptRelation) -> Concepts:
        """Return the related concepts."""
        if relation in get_args(InvertedConceptRelation):
            inverted_relation = inverted(cast(InvertedConceptRelation, relation))
            return tuple(
                concept
                for concept in self.instances.get_all_values()
                if self in concept.get_related_concepts(inverted_relation)
            )
        related_concepts = self.instances.get_values(*self._related_concepts[relation])
        if relation not in get_args(RecursiveConceptRelation):
            return related_concepts
        related_concepts_list = list(related_concepts)
        for concept in related_concepts:
            related_concepts_list.extend(concept.get_related_concepts(relation))
        return tuple(related_concepts_list)

    @property
    def parent(self) -> Concept | None:
        """Return the parent concept."""
        return self.instances.get_values(self._parent)[0] if self._parent else None

    @cached_property
    def base_concept(self) -> Concept:
        """Return the base concept of this concept."""
        return self.parent.base_concept if self.parent else self

    @property
    def constituents(self) -> Concepts:
        """Return the constituent concepts."""
        return self.instances.get_values(*self._constituents)

    def leaf_concepts(self, language: Language) -> Generator[Concept, None, None]:
        """Return this concept's leaf concepts, or self if this concept is a leaf concept."""
        if self.is_composite(language):
            for concept in self.constituents:
                yield from concept.leaf_concepts(language)
        else:
            yield self

    def labels(self, language: Language) -> Labels:
        """Return the labels for the language."""
        if self.is_composite(language):
            return tuple(chain.from_iterable(concept.labels(language) for concept in self.constituents))
        return self._labels.get(language, Labels())

    def colloquial_labels(self, language: Language) -> Labels:
        """Return the colloquial labels for the language."""
        return Labels(label for label in self.labels(language) if label.is_colloquial)

    def non_colloquial_labels(self, language: Language) -> Labels:
        """Return the non-colloquial labels for the language."""
        return Labels(label for label in self.labels(language) if not label.is_colloquial)

    def meanings(self, language: Language) -> Labels:
        """Return the meanings of the concept in the specified language."""
        if self.is_composite(language):
            return tuple(chain.from_iterable(concept.meanings(language) for concept in self.constituents))
        return self._meanings.get(language, Labels())

    def grammatical_categories(self) -> tuple[GrammaticalCategory, ...]:
        """Return the grammatical categories of this concept."""
        keys = self.concept_id.split("/")
        return tuple(cast(GrammaticalCategory, key) for key in keys if key in get_args(GrammaticalCategory))

    def is_composite(self, language: Language) -> bool:
        """Return whether this concept is composite."""
        return self._labels.get(language) is None

    def compounds(self, language: Language) -> Concepts:
        """Return the compounds of the concept."""
        return tuple(concept for concept in self.instances.get_all_values() if self in concept.roots(language))

    def roots(self, language: Language) -> Concepts:
        """Return the root concepts, for the specified language."""
        concept_ids_of_roots = self._roots.get(language, ()) if isinstance(self._roots, dict) else self._roots
        return self.instances.get_values(*concept_ids_of_roots)


Concepts = tuple[Concept, ...]



"""Command to show progress information."""

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Literal

from rich.console import JustifyMethod
from rich.table import Table

from toisto.model.language import Language
from toisto.model.language.iana_language_subtag_registry import ALL_LANGUAGES
from toisto.model.quiz.progress import Progress
from toisto.model.quiz.quiz import Quiz
from toisto.ui.format import format_datetime, format_duration
from toisto.ui.text import console

SortColumn = Literal["attempts", "retention"]


@dataclass(frozen=True)
class QuizSorter:
    """Class to provide a sort function to sort quizzes."""

    progress: Progress
    sort: SortColumn

    def get_sort_key(self, quiz: Quiz) -> int | timedelta:
        """Return the retention attribute to sort by."""
        retention = self.progress.get_retention(quiz)
        if self.sort == "attempts":
            return retention.count
        return retention.length


def show_progress(language: Language, progress: Progress, sort: SortColumn) -> None:
    """Show progress."""
    table = Table(title=f"Progress {ALL_LANGUAGES[language]}")
    justify: dict[str, JustifyMethod] = dict(Attempts="right")
    for column in ("Quiz type", "Question", "From", "To", "Answer(s)", "Attempts", "Retention", "Not quizzed until"):
        table.add_column(column, justify=justify.get(column, "left"))
    sorted_quizzes = sorted(progress.quizzes, key=QuizSorter(progress, sort).get_sort_key, reverse=True)
    for quiz in sorted_quizzes:
        retention = progress.get_retention(quiz)
        skip = retention.skip_until
        quiz_types = " and ".join(quiz.quiz_types).capitalize()
        table.add_row(
            quiz_types,
            quiz.question,
            quiz.question_language,
            quiz.answer_language,
            "\n".join(quiz.answers),
            str(retention.count),
            format_duration(retention.length) if retention.length else "",
            format_datetime(skip) if skip and skip > datetime.now() else "",
        )
    with console.pager():
        console.print(table)



"""Load spelling alternatives."""

import re

from ..metadata import SPELLING_ALTERNATIVES_FILE
from ..model.language import Language
from ..model.language.label import Label
from .json_file import load_json


def load_spelling_alternatives(target_language: Language, source_language: Language) -> None:
    """Load the spelling alternatives."""
    spelling_alternatives = load_json(SPELLING_ALTERNATIVES_FILE)
    key_language_mapping = {
        target_language: target_language,
        source_language: source_language,
        f"{source_language}-if-source-language": source_language,
    }
    for key, language in key_language_mapping.items():
        for regexp, replacement in spelling_alternatives.get(key, {}).items():
            Label.ALTERNATIVES_TO_GENERATE.setdefault(language, {})[re.compile(regexp)] = replacement



"""Utitity functions."""

from collections.abc import Callable, Iterable
from typing import Generic, TypeVar

T = TypeVar("T")


def first(sequence: Iterable[T], where: Callable[[T], bool] = lambda _item: True) -> T:
    """Return the first item in the sequence."""
    return next(item for item in sequence if where(item))


Key = TypeVar("Key")
Value = TypeVar("Value")


class Registry(Generic[Key, Value]):
    """Registry for looking up values by their key."""

    def __init__(self) -> None:
        self.__items: dict[Key, Value] = {}

    def add_item(self, key: Key, value: Value) -> None:
        """Register the item."""
        self.__items[key] = value

    def get_values(self, *keys: Key) -> tuple[Value, ...]:
        """Return the values with the given keys."""
        return tuple(self.__items[key] for key in keys)

    def get_all_values(self) -> tuple[Value, ...]:
        """Return all values."""
        return tuple(self.__items.values())

    def clear(self) -> None:
        """Clear the registry."""
        self.__items.clear()



"""Main module for the application."""

from argparse import Namespace
from configparser import ConfigParser
from contextlib import suppress

with suppress(ImportError):
    import readline  # noqa: F401 `readline` imported but unused

from .command.practice import practice
from .command.show_progress import show_progress
from .metadata import CONCEPT_JSON_FILES, latest_version
from .model.filter import filter_concepts
from .model.quiz.progress import Progress
from .model.quiz.quiz_factory import create_quizzes
from .persistence.config import default_config, read_config
from .persistence.loader import Loader
from .persistence.progress import load_progress
from .persistence.spelling_alternatives import load_spelling_alternatives
from .ui.cli import create_argument_parser, parse_arguments
from .ui.text import console, show_welcome


def init() -> tuple[ConfigParser, Namespace, Progress]:
    """Initialize the main program."""
    argument_parser = create_argument_parser(default_config())
    config = read_config(argument_parser)
    loader = Loader(argument_parser)
    concepts = loader.load(*CONCEPT_JSON_FILES)
    argument_parser = create_argument_parser(config, concepts)
    args = parse_arguments(argument_parser)
    load_spelling_alternatives(args.target_language, args.source_language)
    concepts |= loader.load(*args.file)
    concepts = filter_concepts(concepts, args.concepts, args.target_language, argument_parser)
    quizzes = create_quizzes(args.target_language, args.source_language, *concepts)
    progress = load_progress(args.target_language, quizzes, argument_parser)
    return config, args, progress


def main() -> None:
    """Run the main program."""
    config, args, progress = init()
    match args.command:
        case "progress":
            show_progress(args.target_language, progress, args.sort)
        case _:  # Default command is "practice"
            show_welcome(console.print, latest_version(), config)
            practice(console.print, progress, config)



"""Quiz classes."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import Final, Literal, cast, get_args

from ..language import Language
from ..language.concept import Concept
from ..language.grammar import GrammaticalCategory
from ..language.iana_language_subtag_registry import ALL_LANGUAGES
from ..language.label import Label, Labels
from .match import match

TranslationQuizType = Literal["read", "write"]
ListenQuizType = Literal["dictate", "interpret"]
SemanticQuizType = Literal["answer", "antonym"]
GrammaticalQuizType = Literal[
    "pluralize",
    "singularize",
    "diminutize",
    "masculinize",
    "feminize",
    "neuterize",
    "give positive degree",
    "give comparative degree",
    "give superlative degree",
    "give first person",
    "give second person",
    "give third person",
    "give infinitive",
    "give present tense",
    "give past tense",
    "make declarative",
    "make interrogative",
    "make imperative",
    "affirm",
    "negate",
    "make cardinal",
    "make ordinal",
    "abbreviate",
    "give full form",
]
QuizType = Literal[TranslationQuizType, ListenQuizType, SemanticQuizType, GrammaticalQuizType]
GRAMMATICAL_QUIZ_TYPES: Final[dict[GrammaticalCategory, GrammaticalQuizType]] = {
    "plural": "pluralize",
    "singular": "singularize",
    "diminutive": "diminutize",
    "male": "masculinize",
    "female": "feminize",
    "neuter": "neuterize",
    "positive degree": "give positive degree",
    "comparative degree": "give comparative degree",
    "superlative degree": "give superlative degree",
    "first person": "give first person",
    "second person": "give second person",
    "third person": "give third person",
    "infinitive": "give infinitive",
    "present tense": "give present tense",
    "past tense": "give past tense",
    "declarative": "make declarative",
    "interrogative": "make interrogative",
    "imperative": "make imperative",
    "affirmative": "affirm",
    "negative": "negate",
    "cardinal": "make cardinal",
    "ordinal": "make ordinal",
    "abbreviation": "abbreviate",
    "full form": "give full form",
}
QUIZ_TYPE_GRAMMATICAL_CATEGORIES: Final = {value: key for key, value in GRAMMATICAL_QUIZ_TYPES.items()}
INSTRUCTIONS: Final[dict[Literal[TranslationQuizType, ListenQuizType, SemanticQuizType], str]] = {
    "read": "Translate into",
    "write": "Translate into",
    "dictate": "Listen and write in",
    "interpret": "Listen and write in",
    "answer": "Answer the question in",
    "antonym": "Give the [underline]antonym[/underline] in",
}


@dataclass(frozen=True)
class Quiz:
    """Class representing a quiz."""

    concept: Concept
    question_language: Language
    answer_language: Language
    _question: Label
    _answers: Labels
    quiz_types: tuple[QuizType, ...]
    blocked_by: tuple[Quiz, ...]
    _question_meanings: Labels = ()
    _answer_meanings: Labels = ()

    def __repr__(self) -> str:
        """Return a representation of the quiz for test purposes."""
        return self.key

    def __hash__(self) -> int:
        """Return a hash using the same attributes as used for testing equality."""
        return hash(self.key)

    def __eq__(self, other: object) -> bool:
        """Return whether this quiz is equal to the other."""
        return self.key == other.key if isinstance(other, self.__class__) else False

    def __ne__(self, other: object) -> bool:
        """Return whether this quiz is not equal to the other."""
        return self.key != other.key if isinstance(other, self.__class__) else True

    @cached_property
    def key(self) -> str:
        """Return a string version of the quiz that can be used as key in the progress dict."""
        concept_id = self.concept.base_concept.concept_id
        quiz_types = "+".join(self.quiz_types)
        return f"{concept_id}:{self.question_language}:{self.answer_language}:{self.question}:{quiz_types}"

    def is_correct(self, guess: Label) -> bool:
        """Return whether the guess is correct."""
        return match(guess, *self.answers)

    def is_question(self, guess: Label) -> bool:
        """Return whether the guess is not the answer, but the question (common user error with listening quizzes)."""
        return any(match(guess, *label.spelling_alternatives) for label in (self._question, *self._question_meanings))

    @cached_property
    def question(self) -> Label:
        """Return the first spelling alternative of the question."""
        return self._question.spelling_alternatives[0]

    @property
    def answer(self) -> Label:
        """Return the first spelling alternative of the first answer."""
        return self._answers[0].spelling_alternatives[0]

    @property
    def answers(self) -> Labels:
        """Return all answers."""
        answers = [answer.spelling_alternatives for answer in self._answers]
        return cast(Labels, tuple(chain(*answers)))

    @property
    def non_generated_answers(self) -> Labels:
        """Return all non-generated answers."""
        answers = [answer.non_generated_spelling_alternatives for answer in self._answers]
        return cast(Labels, tuple(chain(*answers)))

    @property
    def question_meanings(self) -> Labels:
        """Return the first spelling alternative of the question meanings."""
        return Labels(meaning.spelling_alternatives[0] for meaning in self._question_meanings)

    @property
    def answer_meanings(self) -> Labels:
        """Return the first spelling alternative of the answer meanings."""
        return Labels(meaning.spelling_alternatives[0] for meaning in self._answer_meanings)

    def other_answers(self, guess: Label) -> Labels:
        """Return the answers not equal to the guess."""
        if self.quiz_types[0] in get_args(ListenQuizType):
            return Labels()  # Returning other answers doesn't make sense if the user has to type what is spoken
        return tuple(
            answer
            for answer in self.non_generated_answers
            if not match(guess, answer) and guess not in answer.spelling_alternatives
        )

    @property
    def instruction(self) -> str:
        """Generate the quiz instruction."""
        if self.is_grammatical:
            categories = " ".join(
                QUIZ_TYPE_GRAMMATICAL_CATEGORIES[cast(GrammaticalQuizType, quiz_type)] for quiz_type in self.quiz_types
            )
            instruction_text = f"Give the [underline]{categories}[/underline] in"
        else:
            quiz_type = cast(Literal[TranslationQuizType, ListenQuizType, SemanticQuizType], self.quiz_types[0])
            if self.question.is_colloquial:
                instruction_text = f"Listen to the colloquial {ALL_LANGUAGES[self.question_language]} and write in"
                if quiz_type == "dictate":
                    instruction_text += " standard"
            else:
                instruction_text = INSTRUCTIONS[quiz_type]
            if self.question.is_complete_sentence:
                instruction_text = instruction_text.replace("write ", "write a complete sentence ")
        return f"{instruction_text} {ALL_LANGUAGES[self.answer_language]}{self._question_note}"

    @property
    def answer_notes(self) -> Sequence[str]:
        """Return the notes to be shown after the quiz has been answered."""
        if "write" in self.quiz_types:
            return tuple(chain.from_iterable(answer.answer_notes for answer in self._answers))
        return self._question.answer_notes

    def is_blocked_by(self, quizzes: Quizzes) -> bool:
        """Return whether this quiz should come after any of the given quizzes."""
        return bool(Quizzes(self.blocked_by) & quizzes)

    @property
    def _question_note(self) -> str:
        """Return the note to be shown as part of the question, if applicable."""
        note_applicable = self.question_language != self.answer_language or {"answer", "dictate"} & set(self.quiz_types)
        question_note = self._answers[0].question_note if "write" in self.quiz_types else self._question.question_note
        return f" ({question_note})" if (note_applicable and question_note) else ""

    @property
    def is_grammatical(self) -> bool:
        """Return whether this is a grammatical quiz."""
        return self.quiz_types[0] in get_args(GrammaticalQuizType)


class Quizzes(set[Quiz]):
    """Set of quizzes."""

    def by_concept(self, concept: Concept) -> Quizzes:
        """Return the quizzes for the concept."""
        return self._quizzes_by_concept.get(concept.base_concept, Quizzes())

    def related_quizzes(self, quiz: Quiz) -> Quizzes:
        """Return the quizzes related to the quiz, meaning quizzes for the same concept and quizzes for examples."""
        quizzes = Quizzes(self.by_concept(quiz.concept))
        for example in quiz.concept.get_related_concepts("example"):
            quizzes |= self.by_concept(example)
        return quizzes

    def by_quiz_type(self, quiz_type: QuizType) -> Quizzes:
        """Return the quizzes of the specified type."""
        return self.__class__(quiz for quiz in self if quiz_type in quiz.quiz_types)

    @cached_property
    def _quizzes_by_concept(self) -> dict[Concept, Quizzes]:
        """Return the quizzes by concept cache.

        Can't use functools.cache as Quizzes instances are not hashable, so use a dict as cache.
        Note that the cache is not updated when quizzes are added or removed after initialization.
        """
        quizzes_by_concept: dict[Concept, Quizzes] = {}
        for quiz in self:
            quizzes_by_concept.setdefault(quiz.concept.base_concept, Quizzes()).add(quiz)
        return quizzes_by_concept



"""Output for the user."""

import sys
from collections.abc import Callable, Sequence
from configparser import ConfigParser
from typing import Final

from rich.console import Console
from rich.panel import Panel

from ..metadata import CHANGELOG_URL, NAME, README_URL, VERSION
from ..model.language.label import END_OF_SENTENCE_PUNCTUATION, Label
from ..model.quiz.quiz import Quiz
from .dictionary import DICTIONARY_URL, linkify_and_enumerate
from .diff import colored_diff
from .style import QUIZ, SECONDARY

console = Console()

LINK_KEY: Final = "⌘ (the command key)" if sys.platform == "darwin" else "Ctrl (the control key)"

WELCOME: Final = f"""👋 Welcome to [underline]{NAME} [white not bold]v{VERSION}[/white not bold][/underline]!

Practice as many words and phrases as you like, for as long as you like.

[{SECONDARY}]{NAME} quizzes you on words and phrases repeatedly. Each time you answer
a quiz correctly, {NAME} will wait longer before repeating it. If you
answer incorrectly, you get one additional attempt to give the correct
answer. If the second attempt is not correct either, {NAME} will reset
the quiz interval.

How does it work?
● To answer a quiz: type the answer, followed by Enter.
● To repeat the spoken text: type Enter without answer.
● To skip to the answer immediately: type ?, followed by Enter.
● To read more about an [link={DICTIONARY_URL}/underlined]underlined[/link] word: keep {LINK_KEY} pressed
  while clicking the word. Not all terminals may support this.
● To quit: type Ctrl-C or Ctrl-D.
[/{SECONDARY}]"""

NEWS: Final = (
    f"🎉 {NAME} [white not bold]{{0}}[/white not bold] is [link={CHANGELOG_URL}]available[/link]. "
    f"Upgrade with [code]pipx upgrade {NAME}[/code]."
)

CONFIG_LANGUAGE_TIP = (
    "️️👉 You may want to use a configuration file to store your language preferences.\n"
    f"See {README_URL.replace('#toisto', '#how-to-configure-toisto')}."
)

DONE: Final = f"""👍 Good job. You're done for now. Please come back later or try a different concept.
[{SECONDARY}]Type `{NAME.lower()} -h` for more information.[/{SECONDARY}]
"""

TRY_AGAIN: Final = "⚠️  Incorrect. Please try again."

TRY_AGAIN_IN_ANSWER_LANGUAGE: Final = "⚠️  Incorrect. Please try again, in [yellow][bold]%(language)s[/bold][/yellow]."

CORRECT: Final = "✅ Correct.\n"

INCORRECT: Final = "❌ Incorrect. "


def feedback_correct(guess: Label, quiz: Quiz) -> str:
    """Return the feedback about a correct result."""
    return CORRECT + meaning(quiz) + other_answers(guess, quiz) + answer_notes(quiz) + examples(quiz)


def feedback_incorrect(guess: Label, quiz: Quiz) -> str:
    """Return the feedback about an incorrect result."""
    if guess == Label(quiz.answer_language, "?"):
        answers = quiz.non_generated_answers
        label = "The correct answer is" if len(answers) == 1 else "The correct answers are"
        feedback = f"{label} {linkify_and_enumerate(*answers)}.\n" + meaning(quiz)
    else:
        label = f'{INCORRECT}The correct answer is "{colored_diff(guess, quiz.answer)}".\n'
        feedback = label + meaning(quiz) + other_answers(quiz.answer, quiz)
    return feedback + answer_notes(quiz)


def meaning(quiz: Quiz) -> str:
    """Return the quiz's meaning, if any."""
    if quiz.question_meanings and quiz.answer_meanings:
        question_meanings = linkify_and_enumerate(*quiz.question_meanings)
        answer_meanings = linkify_and_enumerate(*quiz.answer_meanings)
        meanings = f"{question_meanings}, respectively {answer_meanings}"
    else:
        meanings = linkify_and_enumerate(*(quiz.question_meanings + quiz.answer_meanings))
    return f"[{SECONDARY}]Meaning {meanings}.[/{SECONDARY}]\n" if meanings else ""


def other_answers(guess: Label, quiz: Quiz) -> str:
    """Return the quiz's other answers, if any."""
    if answers := quiz.other_answers(guess):
        label = "Another correct answer is" if len(answers) == 1 else "Other correct answers are"
        return f"""[{SECONDARY}]{label} {linkify_and_enumerate(*answers)}.[/{SECONDARY}]\n"""
    return ""


def examples(quiz: Quiz) -> str:
    """Return the quiz's examples, if any."""
    examples: list[Label] = []
    for example in quiz.concept.get_related_concepts("example"):
        examples.extend(example.labels(quiz.question_language))
    return bulleted_list("Example", examples)


def answer_notes(quiz: Quiz) -> str:
    """Return the answer notes, if any."""
    return bulleted_list("Note", quiz.answer_notes)


def instruction(quiz: Quiz) -> str:
    """Return the instruction for the quiz."""
    return f"[{QUIZ}]{quiz.instruction}:[/{QUIZ}]"


def show_welcome(write_output: Callable[..., None], latest_version: str | None, config: ConfigParser) -> None:
    """Show the welcome message."""
    write_output(WELCOME)
    if latest_version and latest_version.strip("v") > VERSION:
        write_output(Panel(NEWS.format(latest_version), expand=False))
        write_output()
    elif not config.has_section("languages"):
        write_output(Panel(CONFIG_LANGUAGE_TIP, expand=False))
        write_output()


def bulleted_list(label: str, items: Sequence[str], style: str = SECONDARY, bullet: str = "-") -> str:
    """Create a bulleted list of the items."""
    items = [item if item[-1] in END_OF_SENTENCE_PUNCTUATION else f"{item}." for item in items]
    if len(items) == 0:
        return ""
    if len(items) == 1:
        return f"[{style}]{label}: {items[0]}[/{style}]\n"
    return f"[{style}]{label}s:\n" + "\n".join([f"{bullet} {item}" for item in items]) + f"[/{style}]\n"



"""Identifier registry."""

from argparse import ArgumentParser
from pathlib import Path
from typing import Generic, TypeVar

from ..metadata import NAME

Identifier = TypeVar("Identifier")


class IdentifierRegistry(Generic[Identifier]):
    """Registry to check the uniqueness of domain object identifiers across files."""

    def __init__(self, domain_object_name: str, argument_parser: ArgumentParser) -> None:
        self.domain_object_name = domain_object_name
        self.argument_parser = argument_parser
        self.files_by_id: dict[Identifier, Path] = {}

    def check_and_register_identifiers(self, identifiers: tuple[Identifier, ...], file_path: Path) -> None:
        """Check and register the identifiers."""
        for identifier in identifiers:
            self._check_and_register_identifier(identifier, file_path)

    def _check_and_register_identifier(self, identifier: Identifier, file_path: Path) -> None:
        """Check that the identifier is unique and if so, register it. Otherwise, exit with an error message."""
        if identifier in self.files_by_id:
            other_file_path = self.files_by_id[identifier]
            occurs = "occurs multiple times" if file_path == other_file_path else "also occurs"
            name = self.domain_object_name
            self.argument_parser.error(
                f"{NAME} cannot read file {file_path}: {name} identifier '{identifier}' {occurs} in file "
                f"{other_file_path}.\n{name.capitalize()} identifiers must be unique across files.\n",
            )
        self.files_by_id[identifier] = file_path



"""IANA language subtag registry loader.

Toisto includes the IANA language subtag registry located at https://www.iana.org/assignments/language-subtag-registry.
This allows Toisto to check language command line arguments for correctness.
"""

from pathlib import Path

from toisto.metadata import LANGUAGES_FILE
from toisto.model.language import Language
from toisto.tools import first


def load_languages() -> dict[Language, str]:
    """Load the languages from the IANA language subtag registry."""
    languages: dict[Language, str] = {}
    record_lines: list[str] = []
    with Path(LANGUAGES_FILE).open(encoding="utf-8") as language_registry:
        for line in language_registry.readlines():
            if line.startswith("%%"):
                if "Type: language\n" in record_lines:
                    languages.update(parse_record(record_lines))
                record_lines = []  # Reset the lines for the next record
            elif line.startswith("  "):  # Continuation line
                record_lines[-1] += line[1:]  # Skip one space of the continuation
            else:
                record_lines.append(line)
    return languages


def parse_record(lines: list[str]) -> dict[Language, str]:
    """Parse a language record."""
    language = parse_attribute(lines, "Subtag")
    description = parse_attribute(lines, "Description")
    return {Language(language): description}


def parse_attribute(lines: list[str], attribute: str) -> str:
    """Parse an attribute from a language record and return the value."""
    return first(lines, lambda line: line.startswith(attribute)).split(": ", maxsplit=1)[1].strip()



"""Command-line interface."""

import sys
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from configparser import ConfigParser
from pathlib import Path
from typing import get_args

from rich_argparse import RichHelpFormatter

from ..command.show_progress import SortColumn
from ..metadata import BUILT_IN_LANGUAGES, README_URL, SUMMARY, VERSION, latest_version
from ..model.language.concept import Concept
from ..model.language.iana_language_subtag_registry import ALL_LANGUAGES, IANA_LANGUAGE_SUBTAG_REGISTRY_URL


def add_language_arguments(parser: ArgumentParser, config: ConfigParser) -> None:
    """Add the language arguments to the parser."""
    languages = ", ".join(sorted(BUILT_IN_LANGUAGES))
    for argument in ("target", "source"):
        default = config.get("languages", argument, fallback=None)
        default_help = f"default: {default}; " if default else ""
        parser.add_argument(
            f"-{argument[0]}",
            f"--{argument}",
            default=default,
            dest=f"{argument}_language",
            help=f"{argument} language; {default_help}languages available in built-in concepts: {languages}",
            metavar="{language}",
            required=not default,
            type=check_language,
        )


def check_language(language: str) -> str:
    """Check that the language is valid."""
    if language in ALL_LANGUAGES:
        return language
    message = f"invalid choice: '{language}' (see {IANA_LANGUAGE_SUBTAG_REGISTRY_URL} for valid choices)"
    raise ArgumentTypeError(message)


def add_concept_argument(parser: ArgumentParser, concepts: set[Concept]) -> None:
    """Add the concept argument."""
    concept_ids = sorted(concept.concept_id for concept in concepts)
    parser.add_argument(
        "concepts",
        metavar="{concept}",
        nargs="*",
        help=f"concept to use, can be repeated; default: all; built-in concepts: {', '.join(concept_ids)}",
    )


def add_file_arguments(parser: ArgumentParser) -> None:
    """Add the file arguments."""
    parser.add_argument(
        "-f",
        "--file",
        action="append",
        default=[],
        metavar="{file}",
        help="file with extra concepts to read, can be repeated",
        type=Path,
    )


class CommandBuilder:
    """Add commands to the argument parser."""

    def __init__(
        self,
        argument_parser: ArgumentParser,
        concepts: set[Concept],
        config: ConfigParser,
    ) -> None:
        command_help = "default: practice; type `%(prog)s {command} --help` for more information on a command"
        self.subparsers = argument_parser.add_subparsers(dest="command", title="commands", help=command_help)
        self.concepts = concepts
        self.config = config

    def add_command(self, command: str, description: str, command_help: str) -> ArgumentParser:
        """Add a command."""
        parser = self.subparsers.add_parser(
            command,
            description=description,
            help=command_help,
            formatter_class=RichHelpFormatter,
        )
        add_language_arguments(parser, self.config)
        add_file_arguments(parser)
        add_concept_argument(parser, self.concepts)
        return parser

    def add_practice_command(self) -> None:
        """Add a practice command."""
        command_help = (
            "practice a language, for example type `%(prog)s practice --target fi --source en` to "
            "practice Finnish from English"
        )
        self.add_command("practice", "Practice a language.", command_help)

    def add_progress_command(self) -> None:
        """Add a command to show progress."""
        command_help = (
            "show progress, for example `%(prog)s progress --target fi --source en` shows progress "
            "on practicing Finnish from English"
        )
        parser = self.add_command("progress", "Show progress.", command_help)
        parser.add_argument(
            "-S",
            "--sort",
            metavar="{option}",
            choices=sorted(get_args(SortColumn)),
            default="retention",
            help="how to sort progress information; default: by retention; available options: %(choices)s",
        )


def create_argument_parser(config: ConfigParser, concepts: set[Concept] | None = None) -> ArgumentParser:
    """Create the argument parser."""
    epilog = f"See {README_URL} for more information."
    argument_parser = ArgumentParser(description=SUMMARY, epilog=epilog, formatter_class=RichHelpFormatter)
    latest = latest_version()
    version = f"v{VERSION}" + (f" ({latest} is available)" if latest and latest.strip("v") > VERSION else "")
    argument_parser.add_argument("-V", "--version", action="version", version=version)
    builder = CommandBuilder(argument_parser, concepts or set(), config)
    builder.add_practice_command()
    builder.add_progress_command()
    if not {"practice", "progress", "-h", "--help", "-V", "--version"} & set(sys.argv):
        sys.argv.insert(1, "practice")  # Insert practice as default subcommand
    return argument_parser


def parse_arguments(argument_parser: ArgumentParser) -> Namespace:
    """Parse and validate the command-line arguments."""
    namespace = argument_parser.parse_args()
    if namespace.target_language == namespace.source_language:
        message = f"target and source language are the same: '{namespace.target_language}' "
        argument_parser.error(message)
    return namespace



"""Practice command."""

from collections.abc import Callable
from configparser import ConfigParser
from typing import get_args

from toisto.model.language.iana_language_subtag_registry import ALL_LANGUAGES
from toisto.model.language.label import Label
from toisto.model.quiz.progress import Progress
from toisto.model.quiz.quiz import ListenQuizType, Quiz
from toisto.persistence.progress import save_progress
from toisto.ui.dictionary import linkify
from toisto.ui.speech import say
from toisto.ui.text import (
    DONE,
    TRY_AGAIN,
    TRY_AGAIN_IN_ANSWER_LANGUAGE,
    feedback_correct,
    feedback_incorrect,
    instruction,
)


def do_quiz_attempt(quiz: Quiz, config: ConfigParser, attempt: int = 1) -> Label:
    """Present the question and get the answer from the user."""
    while True:
        say(quiz.question_language, quiz.question.pronounceable, config, slow=attempt > 1)
        if answer := Label(quiz.answer_language, input("> ").strip()):
            break
        print("\033[F", end="")  # noqa: T201  # Move cursor one line up
    return answer


def evaluate_answer(quiz: Quiz, progress: Progress, answer: Label, attempt: int = 1) -> str:
    """Evaluate the answer and return the user feedback."""
    if quiz.is_correct(answer):
        progress.mark_correct_answer(quiz)
        return feedback_correct(answer, quiz)
    if answer != Label(quiz.answer_language, "?") and attempt == 1:
        if quiz.is_question(answer) and not quiz.is_grammatical:
            return TRY_AGAIN_IN_ANSWER_LANGUAGE % dict(language=ALL_LANGUAGES[quiz.answer_language])
        return TRY_AGAIN
    progress.mark_incorrect_answer(quiz)
    return feedback_incorrect(answer, quiz)


def do_quiz(write_output: Callable[..., None], quiz: Quiz, progress: Progress, config: ConfigParser) -> None:
    """Do one quiz and update the progress."""
    write_output(instruction(quiz))
    if quiz.quiz_types[0] not in get_args(ListenQuizType):
        write_output(linkify(quiz.question))
    for attempt in range(1, 3):
        answer = do_quiz_attempt(quiz, config, attempt)
        feedback = evaluate_answer(quiz, progress, answer, attempt)
        write_output(feedback)
        if quiz.is_correct(answer) or answer == Label(quiz.answer_language, "?"):
            break


def practice(write_output: Callable[..., None], progress: Progress, config: ConfigParser) -> None:
    """Practice a language."""
    try:
        while quiz := progress.next_quiz():
            do_quiz(write_output, quiz, progress, config)
        write_output(DONE)
    except (KeyboardInterrupt, EOFError):
        write_output()  # Make sure the shell prompt is displayed on a new line
    finally:
        save_progress(progress)
